"""A test that subscribes to NumPy arrays.

Uses REQ/REP (on PUB/SUB socket + 1) to synchronize
"""

# -----------------------------------------------------------------------------
#  Copyright (c) 2010 Brian Granger
#
#  Distributed under the terms of the New BSD License.  The full license is in
#  the file LICENSE.BSD, distributed as part of this software.
# -----------------------------------------------------------------------------

from __future__ import annotations

import time
from argparse import ArgumentParser
from typing import Any, cast

import numpy

import zmq


def recv_array(
    socket: zmq.Socket, flags: int = 0, copy: bool = True, track: bool = False
) -> numpy.ndarray | None:
    """recv a numpy array"""
    header = cast(dict[str, Any], socket.recv_json(flags=flags))
    if header.get('done', False):
        return None
    msg = socket.recv(flags=flags, copy=copy, track=track)
    A = numpy.frombuffer(msg, dtype=header['dtype'])  # type: ignore
    return A.reshape(header['shape'])


def main() -> None:
    parser = ArgumentParser()
    parser.add_argument("--url", default="tcp://127.0.0.1:5555")
    args = parser.parse_args()

    ctx = zmq.Context()
    s = ctx.socket(zmq.SUB)
    s.connect(args.url)
    s.subscribe(b'')

    start = time.perf_counter()
    print("Receiving arrays...")
    a = first_array = recv_array(s)
    assert first_array is not None
    array_count = 0
    while a is not None:
        array_count += 1
        a = recv_array(s)
    print("   Done.")

    end = time.perf_counter()

    elapsed = end - start

    throughput = float(array_count) / elapsed

    message_size = first_array.nbytes
    megabits = float(throughput * message_size * 8) / 1000000

    print(f"message size: {message_size:.0f} [B]")
    print(f"array count: {array_count:.0f}")
    print(f"mean throughput: {throughput:.0f} [msg/s]")
    print(f"mean throughput: {megabits:.3f} [Mb/s]")


if __name__ == "__main__":
    main()



# Copyright (C) PyZMQ Developers
# Distributed under the terms of the Modified BSD License.

import time

import pytest

import zmq
from zmq_test_utils import BaseZMQTestCase, skip_pypy

pytestmark = pytest.mark.skipif(not zmq.DRAFT_API, reason="draft api unavailable")


class TestDraftSockets(BaseZMQTestCase):
    @skip_pypy
    def test_client_server(self):
        client, server = self.create_bound_pair(zmq.CLIENT, zmq.SERVER)
        client.send(b'request')
        msg = self.recv(server, copy=False)
        assert msg.routing_id is not None
        server.send(b'reply', routing_id=msg.routing_id)
        reply = self.recv(client)
        assert reply == b'reply'

    def test_client_server_frame(self):
        client, server = self.create_bound_pair(zmq.CLIENT, zmq.SERVER)
        client.send(b'request')
        msg = self.recv(server, copy=False)
        server.send(msg)
        reply = self.recv(client)
        assert reply == b'request'

    @skip_pypy
    def test_radio_dish(self):
        dish, radio = self.create_bound_pair(zmq.DISH, zmq.RADIO)
        dish.rcvtimeo = 250
        group = 'mygroup'
        dish.join(group)
        received_count = 0
        received = set()
        sent = set()
        for i in range(10):
            msg = str(i).encode('ascii')
            sent.add(msg)
            radio.send(msg, group=group)
            try:
                recvd = dish.recv()
            except zmq.Again:
                time.sleep(0.1)
            else:
                received.add(recvd)
                received_count += 1
        # assert that we got *something*
        assert len(received.intersection(sent)) >= 5


def test_draft_fd():
    if zmq.zmq_version_info() < (4, 3, 2):
        pytest.skip("requires libzmq 4.3.2 for zmq_poller_fd")
    with zmq.Context() as ctx, ctx.socket(zmq.SERVER) as s:
        fd = s.FD
        assert isinstance(fd, int)
        fd_2 = s.get(zmq.FD)
        assert fd_2 == fd



# Copyright (C) PyZMQ Developers
# Distributed under the terms of the Modified BSD License.


import logging
import time

import zmq
from zmq.log import handlers
from zmq_test_utils import BaseZMQTestCase


class TestPubLog(BaseZMQTestCase):
    iface = 'inproc://zmqlog'
    topic = 'zmq'

    @property
    def logger(self):
        # print dir(self)
        logger = logging.getLogger('zmqtest')
        logger.setLevel(logging.DEBUG)
        return logger

    def connect_handler(self, topic=None):
        topic = self.topic if topic is None else topic
        logger = self.logger
        pub, sub = self.create_bound_pair(zmq.PUB, zmq.SUB)
        handler = handlers.PUBHandler(pub)
        handler.setLevel(logging.DEBUG)
        handler.root_topic = topic
        logger.addHandler(handler)
        sub.setsockopt(zmq.SUBSCRIBE, topic.encode())
        time.sleep(0.1)
        return logger, handler, sub

    def test_init_iface(self):
        logger = self.logger
        ctx = self.context
        handler = handlers.PUBHandler(self.iface)
        assert handler.ctx is not ctx
        self.sockets.append(handler.socket)
        # handler.ctx.term()
        handler = handlers.PUBHandler(self.iface, self.context)
        self.sockets.append(handler.socket)
        assert handler.ctx is ctx
        handler.setLevel(logging.DEBUG)
        handler.root_topic = self.topic
        logger.addHandler(handler)
        sub = ctx.socket(zmq.SUB)
        self.sockets.append(sub)
        sub.setsockopt(zmq.SUBSCRIBE, self.topic.encode())
        sub.connect(self.iface)
        import time

        time.sleep(0.25)
        msg1 = 'message'
        logger.info(msg1)

        (topic, msg2) = sub.recv_multipart()
        assert topic == b'zmq.INFO'
        assert msg2 == (msg1 + "\n").encode("utf8")
        logger.removeHandler(handler)

    def test_init_socket(self):
        pub, sub = self.create_bound_pair(zmq.PUB, zmq.SUB)
        logger = self.logger
        handler = handlers.PUBHandler(pub)
        handler.setLevel(logging.DEBUG)
        handler.root_topic = self.topic
        logger.addHandler(handler)

        assert handler.socket is pub
        assert handler.ctx is pub.context
        assert handler.ctx is self.context
        sub.setsockopt(zmq.SUBSCRIBE, self.topic.encode())
        import time

        time.sleep(0.1)
        msg1 = 'message'
        logger.info(msg1)

        (topic, msg2) = sub.recv_multipart()
        assert topic == b'zmq.INFO'
        assert msg2 == (msg1 + "\n").encode("utf8")
        logger.removeHandler(handler)

    def test_root_topic(self):
        logger, handler, sub = self.connect_handler()
        handler.socket.bind(self.iface)
        sub2 = sub.context.socket(zmq.SUB)
        self.sockets.append(sub2)
        sub2.connect(self.iface)
        sub2.setsockopt(zmq.SUBSCRIBE, b'')
        handler.root_topic = b'twoonly'
        msg1 = 'ignored'
        logger.info(msg1)
        self.assertRaisesErrno(zmq.EAGAIN, sub.recv, zmq.NOBLOCK)
        topic, msg2 = sub2.recv_multipart()
        assert topic == b'twoonly.INFO'
        assert msg2 == (msg1 + '\n').encode()

        logger.removeHandler(handler)

    def test_blank_root_topic(self):
        logger, handler, sub_everything = self.connect_handler()
        sub_everything.setsockopt(zmq.SUBSCRIBE, b'')
        handler.socket.bind(self.iface)
        sub_only_info = sub_everything.context.socket(zmq.SUB)
        self.sockets.append(sub_only_info)
        sub_only_info.connect(self.iface)
        sub_only_info.setsockopt(zmq.SUBSCRIBE, b'INFO')
        handler.setRootTopic(b'')
        msg_debug = 'debug_message'
        logger.debug(msg_debug)
        self.assertRaisesErrno(zmq.EAGAIN, sub_only_info.recv, zmq.NOBLOCK)
        topic, msg_debug_response = sub_everything.recv_multipart()
        assert topic == b'DEBUG'
        msg_info = 'info_message'
        logger.info(msg_info)
        topic, msg_info_response_everything = sub_everything.recv_multipart()
        assert topic == b'INFO'
        topic, msg_info_response_onlyinfo = sub_only_info.recv_multipart()
        assert topic == b'INFO'
        assert msg_info_response_everything == msg_info_response_onlyinfo

        logger.removeHandler(handler)

    def test_unicode_message(self):
        logger, handler, sub = self.connect_handler()
        base_topic = (self.topic + '.INFO').encode()
        for msg, expected in [
            ('hello', [base_topic, b'hello\n']),
            ('héllo', [base_topic, 'héllo\n'.encode()]),
            ('tøpic::héllo', [base_topic + '.tøpic'.encode(), 'héllo\n'.encode()]),
        ]:
            logger.info(msg)
            received = sub.recv_multipart()
            assert received == expected
        logger.removeHandler(handler)

    def test_set_info_formatter_via_property(self):
        logger, handler, sub = self.connect_handler()
        handler.formatters[logging.INFO] = logging.Formatter("%(message)s UNITTEST\n")
        handler.socket.bind(self.iface)
        sub.setsockopt(zmq.SUBSCRIBE, handler.root_topic.encode())
        logger.info('info message')
        topic, msg = sub.recv_multipart()
        assert msg == b'info message UNITTEST\n'
        logger.removeHandler(handler)

    def test_custom_global_formatter(self):
        logger, handler, sub = self.connect_handler()
        formatter = logging.Formatter("UNITTEST %(message)s")
        handler.setFormatter(formatter)
        handler.socket.bind(self.iface)
        sub.setsockopt(zmq.SUBSCRIBE, handler.root_topic.encode())
        logger.info('info message')
        topic, msg = sub.recv_multipart()
        assert msg == b'UNITTEST info message'
        logger.debug('debug message')
        topic, msg = sub.recv_multipart()
        assert msg == b'UNITTEST debug message'
        logger.removeHandler(handler)

    def test_custom_debug_formatter(self):
        logger, handler, sub = self.connect_handler()
        formatter = logging.Formatter("UNITTEST DEBUG %(message)s")
        handler.setFormatter(formatter, logging.DEBUG)
        handler.socket.bind(self.iface)
        sub.setsockopt(zmq.SUBSCRIBE, handler.root_topic.encode())
        logger.info('info message')
        topic, msg = sub.recv_multipart()
        assert msg == b'info message\n'
        logger.debug('debug message')
        topic, msg = sub.recv_multipart()
        assert msg == b'UNITTEST DEBUG debug message'
        logger.removeHandler(handler)

    def test_custom_message_type(self):
        class Message:
            def __init__(self, msg: str):
                self.msg = msg

            def __str__(self) -> str:
                return self.msg

        logger, handler, sub = self.connect_handler()
        msg = "hello"
        logger.info(Message(msg))
        topic, received = sub.recv_multipart()
        assert topic == b'zmq.INFO'
        assert received == b'hello\n'
        logger.removeHandler(handler)



from __future__ import annotations

import asyncio

import zmq
import zmq.asyncio


async def main() -> None:
    ctx = zmq.asyncio.Context()

    # shadow exercise
    sync_ctx: zmq.Context = zmq.Context.shadow(ctx)
    ctx2: zmq.asyncio.Context = zmq.asyncio.Context.shadow(sync_ctx)
    ctx2 = zmq.asyncio.Context(sync_ctx)

    url = "tcp://127.0.0.1:5555"
    pub = ctx.socket(zmq.PUB)
    sub = ctx.socket(zmq.SUB)
    pub.bind(url)
    sub.connect(url)
    sub.subscribe(b"")
    await asyncio.sleep(1)

    # shadow exercise
    sync_sock: zmq.Socket[bytes] = zmq.Socket.shadow(pub)
    s2: zmq.asyncio.Socket = zmq.asyncio.Socket(sync_sock)
    s2 = zmq.asyncio.Socket.from_socket(sync_sock)

    print("sending")
    await pub.send(b"plain")
    await pub.send(b"plain")
    await pub.send_multipart([b"topic", b"Message"])
    await pub.send_multipart([b"topic", b"Message"])
    await pub.send_string("asdf")
    await pub.send_pyobj(123)
    await pub.send_json({"a": "5"})

    print("receiving")
    msg_bytes: bytes = await sub.recv()
    msg_frame: zmq.Frame = await sub.recv(copy=False)
    msg_list: list[bytes] = await sub.recv_multipart()
    msg_frames: list[zmq.Frame] = await sub.recv_multipart(copy=False)
    s: str = await sub.recv_string()
    obj = await sub.recv_pyobj()
    d = await sub.recv_json()

    pub.close()
    sub.close()


if __name__ == "__main__":
    asyncio.run(main())



"""Print a markdown table of sdist/wheel outputs

for use in github job summary
"""

from pathlib import Path


def make_summary(dist_dir: str | Path) -> str:
    """Render a list of files as a markdown table

    For use summarizing wheel outputs
    """

    dist_dir = Path(dist_dir)
    all_dists = sorted(dist_dir.glob("*"))
    lines = [
        f"### {len(all_dists)} files",
        "",
        "| filename | size |",
        "|----------|------|",
    ]
    for path in all_dists:
        size = path.stat().st_size
        if size < 1e6:
            size_s = f"{size / 1e3:.0f} kB"
        else:
            size_s = f"{size / 1e6:.1f} MB"
        lines.append(f"| {path.name} | {size_s} |")
    return "\n".join(lines)


if __name__ == "__main__":
    import sys

    if len(sys.argv) > 1:
        dist_dir = Path(sys.argv[1])
    else:
        dist_dir = Path("dist")
    print(make_summary(dist_dir))



"""
Use recv_into to receive data directly into a numpy array
"""

import numpy as np
import numpy.testing as nt

import zmq

url = "inproc://test"


def main() -> None:
    A = (np.random.random((5, 5)) * 255).astype(dtype=np.int64)
    B = np.empty_like(A)
    assert not (A == B).all()

    with zmq.Context() as ctx, ctx.socket(zmq.PUSH) as push, ctx.socket(
        zmq.PULL
    ) as pull:
        push.bind(url)
        pull.connect(url)
        print("sending:\n", A)
        push.send(A)
        bytes_received = pull.recv_into(B)
        print(f"received {bytes_received} bytes:\n", B)
        assert bytes_received == A.nbytes
        nt.assert_allclose(A, B)


if __name__ == "__main__":
    main()



# Copyright (C) PyZMQ Developers
# Distributed under the terms of the Modified BSD License.


import copy
import gc

try:
    from sys import getrefcount
except ImportError:
    grc = None
else:
    grc = getrefcount

import time

import pytest

import zmq
from zmq.utils.garbage import gc as zmq_gc
from zmq_test_utils import PYPY, BaseZMQTestCase, SkipTest, skip_cpython_cffi, skip_pypy

# some useful constants:

x = b'x'

if grc:
    rc0 = grc(x)
    v = memoryview(x)
    view_rc = grc(x) - rc0


def await_gc(gc_key):
    """wait for zmq garbage collection

    Necessary because of the zero-copy gc thread,
    which can take some time to receive its DECREF message.
    """
    deadline = time.monotonic() + 3
    gc.collect()
    while time.monotonic() < deadline:
        if gc_key in zmq_gc.refs:
            time.sleep(0.05)
        else:
            gc.collect()
            return
    raise TimeoutError("gc not collected")


class TestFrame(BaseZMQTestCase):
    def setUp(self):
        super().setUp()
        # make sure we are starting clean
        assert not zmq_gc.refs

    def tearDown(self):
        super().tearDown()
        for i in range(3):
            gc.collect()

        deadline = time.monotonic() + 3
        while zmq_gc.refs and time.monotonic() < deadline:
            time.sleep(0.05)

        # make sure we left no refs
        assert not zmq_gc.refs

    @skip_pypy
    def test_above_30(self):
        """Message above 30 bytes are never copied by 0MQ."""
        for i in range(5, 16):  # 32, 64,..., 65536
            s = (2**i) * x
            rc = grc(s)
            m = zmq.Frame(s, copy=False)
            _gc_ref = next(iter(zmq_gc.refs))
            assert grc(s) == rc + 2
            del m
            await_gc(_gc_ref)
            assert grc(s) == rc
            del s

    def test_str(self):
        """Test the str representations of the Frames."""
        m = zmq.Frame(b'')
        assert str(m) == "<zmq.Frame(b'')>"
        m = zmq.Frame(b'123456789')
        assert str(m) == "<zmq.Frame(b'123456789')>"
        m = zmq.Frame(b'x' * 20)
        assert str(m) == "<zmq.Frame(b'xxxxxxxxxxxx'...20B)>"
        m = zmq.Frame(b'x' * 2000)
        assert str(m) == "<zmq.Frame(b'xxxxxxxxxxxx'...2kB)>"
        m = zmq.Frame(b'x' * 2_000_000)
        assert str(m) == "<zmq.Frame(b'xxxxxxxxxxxx'...2MB)>"

    def test_bytes(self):
        """Test the Frame.bytes property."""
        for i in range(1, 16):
            s = (2**i) * x
            m = zmq.Frame(s)
            b = m.bytes
            assert s == m.bytes
            if not PYPY:
                # check that it copies
                assert b is not s
            # check that it copies only once
            assert b is m.bytes

    def test_unicode(self):
        """Test the unicode representations of the Frames."""
        s = 'asdf'
        self.assertRaises(TypeError, zmq.Frame, s)
        for i in range(16):
            s = (2**i) * '§'
            m = zmq.Frame(s.encode('utf8'))
            assert s == m.bytes.decode('utf8')

    def test_len(self):
        """Test the len of the Frames."""
        for i in range(16):
            s = (2**i) * x
            m = zmq.Frame(s)
            assert len(s) == len(m)

    @skip_pypy
    def test_lifecycle1(self):
        """Run through a ref counting cycle with a copy."""
        for i in range(5, 16):  # 32, 64,..., 65536
            s = (2**i) * x
            rc = rc_0 = grc(s)
            m = zmq.Frame(s, copy=False)
            _gc_ref = next(iter(zmq_gc.refs))
            rc += 2
            assert grc(s) == rc
            m2 = copy.copy(m)
            rc += 1
            assert grc(s) == rc
            # no increase in refcount for accessing buffer
            # which references m2 directly
            buf = m2.buffer
            assert grc(s) == rc

            assert s == bytes(m2)
            assert s == m.bytes
            assert s == bytes(buf)
            del m2
            assert grc(s) == rc
            # buf holds direct reference to m2 which holds
            del buf
            rc -= 1
            assert grc(s) == rc
            del m
            rc -= 2
            await_gc(_gc_ref)
            assert grc(s) == rc
            assert rc == rc_0
            del s

    @skip_pypy
    def test_lifecycle2(self):
        """Run through a different ref counting cycle with a copy."""
        for i in range(5, 16):  # 32, 64,..., 65536
            s = (2**i) * x
            rc = rc_0 = grc(s)
            m = zmq.Frame(s, copy=False)
            _gc_ref = next(iter(zmq_gc.refs))
            rc += 2
            assert grc(s) == rc
            m2 = copy.copy(m)
            rc += 1
            assert grc(s) == rc
            # no increase in refcount for accessing buffer
            # which references m directly
            buf = m.buffer
            assert grc(s) == rc
            assert s == bytes(m2)
            assert s == m2.bytes
            assert s == m.bytes
            assert s == bytes(buf)
            del buf
            assert grc(s) == rc
            del m
            rc -= 1
            assert grc(s) == rc
            del m2
            rc -= 2
            await_gc(_gc_ref)
            assert grc(s) == rc
            assert rc == rc_0
            del s

    def test_tracker(self):
        m = zmq.Frame(b'asdf', copy=False, track=True)
        assert not m.tracker.done
        pm = zmq.MessageTracker(m)
        assert not pm.done
        del m
        for i in range(3):
            gc.collect()
        for i in range(10):
            if pm.done:
                break
            time.sleep(0.1)
        assert pm.done

    def test_no_tracker(self):
        m = zmq.Frame(b'asdf', track=False)
        assert m.tracker is None
        m2 = copy.copy(m)
        assert m2.tracker is None
        self.assertRaises(ValueError, zmq.MessageTracker, m)

    def test_multi_tracker(self):
        m = zmq.Frame(b'asdf', copy=False, track=True)
        m2 = zmq.Frame(b'whoda', copy=False, track=True)
        mt = zmq.MessageTracker(m, m2)
        assert not m.tracker.done
        assert not mt.done
        self.assertRaises(zmq.NotDone, mt.wait, 0.1)
        del m
        for i in range(3):
            gc.collect()
        self.assertRaises(zmq.NotDone, mt.wait, 0.1)
        assert not mt.done
        del m2
        for i in range(3):
            gc.collect()
        assert mt.wait(0.1) is None
        assert mt.done

    def test_buffer_in(self):
        """test using a buffer as input"""
        ins = "§§¶•ªº˜µ¬˚…∆˙åß∂©œ∑´†≈ç√".encode()
        zmq.Frame(memoryview(ins))
        zmq.Frame(bytearray(5))

    def test_bad_buffer_in(self):
        """test using a bad object"""
        with pytest.raises(TypeError):
            zmq.Frame(5)
        with pytest.raises(TypeError):
            zmq.Frame(object())
        with pytest.raises(TypeError):
            zmq.Frame("str")
        with pytest.raises(BufferError):
            zmq.Frame(memoryview(bytearray(10))[::2])

    def test_buffer_out(self):
        """receiving buffered output"""
        ins = "§§¶•ªº˜µ¬˚…∆˙åß∂©œ∑´†≈ç√".encode()
        m = zmq.Frame(ins)
        outb = m.buffer
        assert isinstance(outb, memoryview)
        assert outb is m.buffer
        assert m.buffer is m.buffer

    @skip_cpython_cffi
    def test_memoryview_shape(self):
        """memoryview shape info"""
        data = "§§¶•ªº˜µ¬˚…∆˙åß∂©œ∑´†≈ç√".encode()
        n = len(data)
        f = zmq.Frame(data)
        view1 = f.buffer
        assert view1.ndim == 1
        assert view1.shape == (n,)
        assert view1.tobytes() == data
        view2 = memoryview(f)
        assert view2.ndim == 1
        assert view2.shape == (n,)
        assert view2.tobytes() == data

    def test_multisend(self):
        """ensure that a message remains intact after multiple sends"""
        a, b = self.create_bound_pair(zmq.PAIR, zmq.PAIR)
        s = b"message"
        m = zmq.Frame(s)
        assert s == m.bytes

        a.send(m, copy=False)
        time.sleep(0.1)
        assert s == m.bytes
        a.send(m, copy=False)
        time.sleep(0.1)
        assert s == m.bytes
        a.send(m, copy=True)
        time.sleep(0.1)
        assert s == m.bytes
        a.send(m, copy=True)
        time.sleep(0.1)
        assert s == m.bytes
        for i in range(4):
            r = b.recv()
            assert s == r
        assert s == m.bytes

    def test_memoryview(self):
        """test messages from memoryview"""
        s = b'carrotjuice'
        memoryview(s)
        m = zmq.Frame(s)
        buf = m.buffer
        s2 = buf.tobytes()
        assert s2 == s
        assert m.bytes == s

    @skip_cpython_cffi
    def test_noncopying_recv(self):
        """check for clobbering message buffers"""
        null = b'\0' * 64
        sa, sb = self.create_bound_pair(zmq.PAIR, zmq.PAIR)
        for i in range(32):
            # try a few times
            sb.send(null, copy=False)
            m = sa.recv(copy=False)
            mb = m.bytes
            # buf = memoryview(m)
            buf = m.buffer
            del m
            for i in range(5):
                ff = b'\xff' * (40 + i * 10)
                sb.send(ff, copy=False)
                m2 = sa.recv(copy=False)
                b = buf.tobytes()
                assert b == null
                assert mb == null
                assert m2.bytes == ff
                assert type(m2.bytes) is bytes

    @skip_cpython_cffi
    def test_noncopying_memoryview(self):
        """test non-copying memmoryview messages"""
        null = b'\0' * 64
        sa, sb = self.create_bound_pair(zmq.PAIR, zmq.PAIR)
        for i in range(32):
            # try a few times
            sb.send(memoryview(null), copy=False)
            m = sa.recv(copy=False)
            buf = memoryview(m)
            for i in range(5):
                ff = b'\xff' * (40 + i * 10)
                sb.send(memoryview(ff), copy=False)
                m2 = sa.recv(copy=False)
                buf2 = memoryview(m2)
                assert buf.tobytes() == null
                assert not buf.readonly
                assert buf2.tobytes() == ff
                assert not buf2.readonly
                assert type(buf) is memoryview

    def test_buffer_numpy(self):
        """test non-copying numpy array messages"""
        try:
            import numpy
            from numpy.testing import assert_array_equal
        except ImportError:
            raise SkipTest("requires numpy")
        rand = numpy.random.randint
        shapes = [rand(2, 5) for i in range(5)]
        a, b = self.create_bound_pair(zmq.PAIR, zmq.PAIR)
        dtypes = [int, float, '>i4', 'B']
        for i in range(1, len(shapes) + 1):
            shape = shapes[:i]
            for dt in dtypes:
                A = numpy.empty(shape, dtype=dt)
                a.send(A, copy=False)
                msg = b.recv(copy=False)

                B = numpy.frombuffer(msg, A.dtype).reshape(A.shape)
                assert_array_equal(A, B)

            A = numpy.empty(shape, dtype=[('a', int), ('b', float), ('c', 'a32')])
            A['a'] = 1024
            A['b'] = 1e9
            A['c'] = 'hello there'
            a.send(A, copy=False)
            msg = b.recv(copy=False)

            B = numpy.frombuffer(msg, A.dtype).reshape(A.shape)
            assert_array_equal(A, B)

    @skip_pypy
    def test_frame_more(self):
        """test Frame.more attribute"""
        frame = zmq.Frame(b"hello")
        assert not frame.more
        sa, sb = self.create_bound_pair(zmq.PAIR, zmq.PAIR)
        sa.send_multipart([b'hi', b'there'])
        frame = self.recv(sb, copy=False)
        assert frame.more
        if zmq.zmq_version_info()[0] >= 3 and not PYPY:
            assert frame.get(zmq.MORE)
        frame = self.recv(sb, copy=False)
        assert not frame.more
        if zmq.zmq_version_info()[0] >= 3 and not PYPY:
            assert not frame.get(zmq.MORE)



"""
use recv_into with an empty buffer to discard unwanted message frames

avoids unnecessary allocations for message frames that won't be used
"""

import logging
import os
import random
import secrets
import time
from pathlib import Path
from tempfile import TemporaryDirectory
from threading import Thread

import zmq

EMPTY = bytearray()


def subscriber(url: str) -> None:
    log = logging.getLogger("subscriber")
    with zmq.Context() as ctx, ctx.socket(zmq.SUB) as sub:
        sub.linger = 0
        sub.connect(url)
        sub.subscribe(b"")
        log.info("Receiving...")
        while True:
            frame_0 = sub.recv_string()
            if frame_0 == "exit":
                log.info("Exiting...")
                break
            elif frame_0 == "large":
                discarded_bytes = 0
                discarded_frames = 0
                while sub.rcvmore:
                    discarded_bytes += sub.recv_into(EMPTY)
                    discarded_frames += 1
                log.info(
                    "Discarding large message frames: %s, bytes: %s",
                    discarded_frames,
                    discarded_bytes,
                )
            else:
                msg: list = [frame_0]
                if sub.rcvmore:
                    msg.extend(sub.recv_multipart(flags=zmq.DONTWAIT))
                log.info("Received %s", msg)
    log.info("Done")


def publisher(url) -> None:
    log = logging.getLogger("publisher")
    choices = ["large", "small"]
    with zmq.Context() as ctx, ctx.socket(zmq.PUB) as pub:
        pub.linger = 1000
        pub.bind(url)
        time.sleep(1)
        for i in range(10):
            kind = random.choice(choices)
            frames = [kind.encode()]
            if kind == "large":
                for _ in range(random.randint(0, 5)):
                    chunk_size = random.randint(1024, 2048)
                    chunk = os.urandom(chunk_size)
                    frames.append(chunk)
            else:
                for _ in range(random.randint(0, 3)):
                    chunk_size = random.randint(0, 5)
                    chunk = secrets.token_hex(chunk_size).encode()
                    frames.append(chunk)
            nbytes = sum(len(chunk) for chunk in frames)
            log.info("Sending %s: %s bytes", kind, nbytes)
            pub.send_multipart(frames)
            time.sleep(0.1)
        log.info("Sending exit")
        pub.send(b"exit")
    log.info("Done")


def main() -> None:
    logging.basicConfig(level=logging.INFO)
    with TemporaryDirectory() as td:
        sock_path = Path(td) / "example.sock"
        url = f"ipc://{sock_path}"
        s_thread = Thread(
            target=subscriber, args=(url,), daemon=True, name="subscriber"
        )
        s_thread.start()
        publisher(url)
        s_thread.join(timeout=3)


if __name__ == "__main__":
    main()
